/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* plasma.c
   Code for drawing a nice plasma */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "plasma.h"
#include "cursor.h"

#define T_PLASMA_SIZE 64
#define T_PLASMA_2SIZE (T_PLASMA_SIZE/2)

/* texture for each plasma */
GLuint t_plasma = 0;
unsigned char *circle_raw = NULL;

/* position of plasmas */
float plasma_pos[11][2];
/* color of plasmas */
float plasma_col[11][3];
/* depl. of plasmas */
float plasma_moves[11][2];
/* scale of plasmas */
float plasma_sizes[11][2];
float plasma_spesize[2];
/* color variations of plasmas */
float plasma_dir[11][3];


/* regenerate plasma textures */
void regen_plasma(World * world)
{/* a little bit hard! */
  init_plasma(world);
}


/* free all plasma relative stuff */
void free_plasma(World *world)
{
  /* Remove existing texture */
  if (t_plasma != 0)
    glDeleteTextures(1, &t_plasma);
}


/* init the corresponding data */
void init_plasma(World *world)
{
  int i, j;
  float d;

  /* Remove existing texture */
  if (t_plasma != 0)
    glDeleteTextures(1, &t_plasma);

  circle_raw = malloc(sizeof(unsigned char)*T_PLASMA_SIZE*T_PLASMA_SIZE*4);
  if (circle_raw == NULL)
    {
    Neprintf("cannot allocate %d bytes in 'init_plasma'. Plasma aborted.\n", sizeof(unsigned char)*T_PLASMA_SIZE*T_PLASMA_SIZE*4);
    return;
    }

  for(i=0; i<T_PLASMA_SIZE; i++)
    for(j=0; j<T_PLASMA_SIZE; j++)
      {
      circle_raw[(i+j*T_PLASMA_SIZE)*4+0] = 255;
      circle_raw[(i+j*T_PLASMA_SIZE)*4+1] = 255;
      circle_raw[(i+j*T_PLASMA_SIZE)*4+2] = 255;
      d = sqrt(nSQR(T_PLASMA_2SIZE-(i+.5))+nSQR(T_PLASMA_2SIZE-(j+.5)));
      if (d > 9.*T_PLASMA_2SIZE/10.)
        circle_raw[(i+j*T_PLASMA_SIZE)*4+3] = 0;
      else
        circle_raw[(i+j*T_PLASMA_SIZE)*4+3] = (unsigned char)(255.*.7*(1. - d*10./(9.*T_PLASMA_2SIZE)));
      }

  /* Generate the box texture */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &t_plasma);
  glBindTexture(GL_TEXTURE_2D, t_plasma);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, world->texture_mode);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, world->texture_mode);
  glTexImage2D(GL_TEXTURE_2D,
	       0,
	       4,
	       T_PLASMA_SIZE,
	       T_PLASMA_SIZE,
	       0,
	       GL_RGBA,
	       GL_UNSIGNED_BYTE,
	       (const GLvoid*)circle_raw);

  free(circle_raw);
  circle_raw = NULL;

  /* init datas */
  for(i=0; i<11; i++)
    {
    plasma_sizes[i][0] = random_float(45., 95.);
    plasma_sizes[i][1] = random_float(45., 95.);
    
    plasma_pos[i][0] = random_float(-40, 70);
    plasma_pos[i][1] = random_float(-40, 70);
    
    plasma_col[i][0] = random_float(0.3, 1.);
    plasma_col[i][1] = random_float(0.3, 1.);
    plasma_col[i][2] = random_float(0.3, 1.);
    
    plasma_moves[i][0] = random_float(-6, 6);
    plasma_moves[i][1] = random_float(-6, 6);
    
    if (random_1on2())
      plasma_dir[i][0] = 1;
    else
      plasma_dir[i][0] = -1;
    if (random_1on2())
      plasma_dir[i][1] = 1;
    else
      plasma_dir[i][1] = -1;
    if (random_1on2())
      plasma_dir[i][2] = 1;
    else
      plasma_dir[i][2] = -1;
    }
  plasma_spesize[0] = plasma_sizes[0][0];
  plasma_spesize[1] = plasma_sizes[0][1];
}

/* draw a plasma (suppose that 2D mode is anabled) */
void draw_plasma(World *world)
{
  static float last_time = 0.;
  float dt;
  int nb, i, x, y;

  if (t_plasma == 0)
    return;

  nb = world->plasma_detail + 4;
  if (nb < 4)
    nb = 4;
  if (nb > 11)
    nb = 11;

  x = get_cursor_x();
  y = get_cursor_y();

  /* draw the plasma */
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
  glBindTexture(GL_TEXTURE_2D, t_plasma);

  glBegin(GL_QUADS);

  plasma_pos[0][0] = x - plasma_sizes[0][0]/2.;
  plasma_pos[0][1] = y - plasma_sizes[0][1]/2.;
  plasma_sizes[0][0] = plasma_spesize[0] + 25. *
                sin(world->total_elapsed_time*2.*M_PI/4.);
  plasma_sizes[0][1] = plasma_spesize[1] + 25. *
                sin(world->total_elapsed_time*2.*M_PI/4.);
  for(i=0; i<nb; i++)
    {
    glColor4f(plasma_col[i][0],plasma_col[i][1],plasma_col[i][2],1.);
    glTexCoord2f(0,0);
    glVertex2f(plasma_pos[i][0], plasma_pos[i][1]);
    glTexCoord2f(1,0);
    glVertex2f(plasma_pos[i][0]+plasma_sizes[i][0], plasma_pos[i][1]);
    glTexCoord2f(1,1);
    glVertex2f(plasma_pos[i][0]+plasma_sizes[i][0], plasma_pos[i][1]+plasma_sizes[i][1]);
    glTexCoord2f(0,1);
    glVertex2f(plasma_pos[i][0], plasma_pos[i][1]+plasma_sizes[i][1]);
    }

  glEnd();

  dt = world->total_elapsed_time - last_time;
  last_time = world->total_elapsed_time;
  if (dt > 1.)
    dt = 1.;

  for(i=0; i<nb; i++)
    {
    plasma_pos[i][0] += plasma_moves[i][0] * dt;
    plasma_pos[i][1] += plasma_moves[i][1] * dt;
    
    if (plasma_pos[i][0]+plasma_sizes[i][0]/2. > 99.)
      {
      plasma_pos[i][0] = 98. - plasma_sizes[i][0]/2.;
      plasma_moves[i][0] = -plasma_moves[i][0];
      }
    if (plasma_pos[i][0] < -plasma_sizes[i][0]/2.)
      {
      plasma_pos[i][0] = -plasma_sizes[i][0]/2. + 1;
      plasma_moves[i][0] = -plasma_moves[i][0];
      }
    if (plasma_pos[i][1]+plasma_sizes[i][1]/2. > 99.)
      {
      plasma_pos[i][1] = 98. - plasma_sizes[i][1]/2.;
      plasma_moves[i][1] = -plasma_moves[i][1];
      }
    if (plasma_pos[i][1] < -plasma_sizes[i][1]/2.)
      {
      plasma_pos[i][1] = -plasma_sizes[i][1]/2. + 1;
      plasma_moves[i][1] = -plasma_moves[i][1];
      }
    /* update colors */
    plasma_col[i][0] += dt * 0.1 * plasma_dir[i][0];
    plasma_col[i][1] += dt * 0.1 * plasma_dir[i][1];
    plasma_col[i][2] += dt * 0.1 * plasma_dir[i][2];
    if ((plasma_dir[i][0] > 0)&&(plasma_col[i][0] > 1.))
      {
      plasma_col[i][0] = 1.;	
      if (random_1on2())
	plasma_dir[i][0] = 1;
      else
	plasma_dir[i][0] = -1;
      }
    if ((plasma_dir[i][0] < 0)&&(plasma_col[i][0] < .2))
      {
      plasma_col[i][0] = .2;	
      if (random_1on2())
	plasma_dir[i][0] = 1;
      else
	plasma_dir[i][0] = -1;
      }
    if ((plasma_dir[i][1] > 0)&&(plasma_col[i][1] > 1.))
      {
      plasma_col[i][1] = 1.;	
      if (random_1on2())
	plasma_dir[i][1] = 1;
      else
	plasma_dir[i][1] = -1;
      }
    if ((plasma_dir[i][1] < 0)&&(plasma_col[i][1] < .2))
      {
      plasma_col[i][1] = .2;	
      if (random_1on2())
	plasma_dir[i][1] = 1;
      else
	plasma_dir[i][1] = -1;
      }
    if ((plasma_dir[i][2] > 0)&&(plasma_col[i][2] > 1.))
      {
      plasma_col[i][2] = 1.;	
      if (random_1on2())
	plasma_dir[i][2] = 1;
      else
	plasma_dir[i][2] = -1;
      }
    if ((plasma_dir[i][2] < 0)&&(plasma_col[i][2] < .2))
      {
      plasma_col[i][2] = .2;	
      if (random_1on2())
	plasma_dir[i][2] = 1;
      else
	plasma_dir[i][2] = -1;
      }
    }
}


