/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* player.c
   This file contains all the player related code */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#ifdef USE_GLASS
#include <glass.h>
#endif

#include "player.h"

#include "world.h"
#include "game.h"
#include "round.h"
#include "turn.h"

#include "text.h"
#include "vector.h"
#include "map.h"
#include "tank.h"
#include "projectile.h"
#include "2d.h"
#include "draw.h"
#include "ai.h"
#include "tools.h"
#include "display_stack.h"
#include "glut_interface.h"

static Player *allocate_player(void);
static void build_weapon_list(Player *player);
static Weapon *find_player_weapon(Player *player, const char *weapon);
static int clear_weapon_list(Player *player);
static void update_player_direction(Player *player);
static void update_player_directionv(Player *player);
static void update_player(Player *player, float dt);
static void move_player(Player *player, float dt);
static void update_player_finances(Player *player, float dt);
static void free_player(Player *player);

/* Allocate space for a new player */
static Player *allocate_player(void)
{
  Player *player = malloc(sizeof(Player));
  int i;

  if(!player)
  {
    Neprintf("Error: Unable to allocate space for player\n");
    return NULL;
  }

  player->world = NULL;
  player->game = NULL;

  player->name = NULL;
  player->cash = 0;
  player->winnings = 0;
  player->temp_cash = 0;
  player->temp_winnings = 0;
  player->percent_transferred = 0.0;
  player->cash_display_timer = 0.0;
  player->health = 100.0;

  player->round_winnings = 0;
  player->round_cost = 0;

  player->round_kills = 0;
  player->kills = 0;

  player->player_stats = NULL;

  player->player_state = PLAYER_DEAD;
  player->player_type = PLAYER_HUMAN;
  player->ai = NULL;

  player->weapon_list = NULL;
  player->selected_weapon = NULL;

  player->tank = NULL;
  player->tank_state = IN_AIR;

  player->engine_on = FALSE;
  player->has_driven = FALSE;
  player->brakes = FALSE;
  player->speed = 0.0;
  player->accel = 0.0;

  player->underwater_time = 0.0;

  for(i = 0; i < 3; i++)
  {
    player->pos[i] = 0.0;
    player->dir[i] = 0.0;
    player->vel[i] = 0.0;
    player->acc[i] = 0.0;
    player->rot_dir[i] = 0.0;
  }
  /* Point in a valid direction */
  player->dir[2] = 1.0;
  /* To stop error when converting vel[] to dir[] */
  player->vel[1] = -0.01;

  player->rot_dir[0] = 1.0;
  player->rot_angle = 0.0;
  player->heading = 0.0;

  /* Always in the worlds gravity */
  player->acc[1] = -9.81;

  player->turret.heading = 0.0;
  player->turret.pitch = -30.0;
  player->muzzle_vel = 25.0;
  player->thrust = 30.0;

  player->next = NULL;
  player->last = NULL;

  return player;
}

/* Add a new player into the world */
Player *add_player(World *world, const char *name)
{
  Player *player, **player_ptr = &world->player_list, *last_player = NULL;

  player = allocate_player();
  if(!player)
    return NULL;

  player->world = world;
  set_player_name(player, name);

  /* Build up the weapon list for this player, based on the availiable weapons
     in the world */
  build_weapon_list(player);

  /* Add the player onto the end of the list */
  while(*player_ptr != NULL)
  {
    last_player = *player_ptr;
    player_ptr = &(*player_ptr)->next;
  }
  *player_ptr = player;
  player->last = last_player;

  Nprintf("Player \"%s\" added\n", player->name);

  /* Give the player the first tank */
  player->tank = world->tank_list;
  if(player->tank == NULL)
    Neprintf("Warning: player added, but no tanks present\n");

  /* Update everyones player statistics */
  build_all_player_statistics(world);

  return player;
}

/* Set the players name */
void set_player_name(Player *player, const char *name)
{
  Nfree(player->name);
  if(name == NULL)
    player->name = strdup("");
  else
    player->name = strdup(name);
}

/* Set the amount of cash the player has */
void set_player_cash(Player *player, int cash)
{
  player->cash = cash;
}

/* Give the player some winnings */
void give_player_winnings(Player *player, int winnings)
#define FADE_IN_TIME 2.0
{
  /* Don't bother if they didn't win anything */
  if(winnings <= 0)
    return;

  /* Set the cash display timer to 0, or if already counting
     winnings, just after the display has faded up */
  if(player->winnings)
  {
    player->percent_transferred *= player->winnings / (player->winnings + winnings);
    if(player->cash_display_timer == 0.0)
      player->cash_display_timer = FADE_IN_TIME;
  }
  else
    player->cash_display_timer = 0.0000001;
  /* (set to 0.0001 because timer not active when = 0.0) */

  player->winnings += winnings;
  player->temp_cash = player->cash;
  player->temp_winnings = player->winnings;
}

/* Set the type of the player (human, ai...) */
void set_player_type(Player *player, PlayerType type)
{
  /* No need to change */
  if(player->player_type == type)
    return;

  player->player_type = type;

  /* if AI, prepare its data */
  if((type == PLAYER_HUMAN)||(type == PLAYER_HUMAN_NET))
  {
    free_ai(player->ai);
    player->ai = NULL;
    return;
  }

  make_ai(player);
}

/* Set the players tank, return FALSE if tank not found */
int set_player_tank(Player *player, const char *tank_name)
{
  Tank *tank = find_tank(player->world, tank_name);

  player->tank = tank;

  if(!tank)
  {
    Neprintf("  Unable to find tank \"%s\", for player \"%s\".\n", tank_name, player->name);
    return FALSE;
  }
  else
    return TRUE;
}

/* Build up the weapon list for this player, based on the availiable weapons
   in the world */
static void build_weapon_list(Player *player)
{
  World *world = player->world;
  Munition *mun_ptr = world->munition_list;
  Weapon *weapon, **weap_ptr = &player->weapon_list, *old_weapon = NULL;

  /* If no munitions */
  if(world->munition_list == NULL)
    return;

  /* Go through the munitions listed in the world */
  do
  {
    /* If the weapon can be bought/used */
    if(mun_ptr->usable)
    {
      /* Make a new weapon entry */
      *weap_ptr = malloc(sizeof(Weapon));
      weapon = *weap_ptr;

      /* Fill the data, and add to the list */
      weapon->munition = mun_ptr;
      weapon->number = 0;
      weapon->last = old_weapon;
      if(old_weapon != NULL)
	weapon->last->next = weapon;
      weapon->next = NULL;

      old_weapon = weapon;
      weap_ptr = &(*weap_ptr)->next;
    }

    mun_ptr = mun_ptr->next;
  } while(mun_ptr != world->munition_list);

  /* Select the first weapon (should be the null weapon) */
  player->selected_weapon = player->weapon_list;
}

/* Find the weapon named `weapon' a player has, return NULL if not found */
static Weapon *find_player_weapon(Player *player, const char *weapon)
{
  Weapon *weap_ptr = player->weapon_list;

  while(weap_ptr != NULL)
  {
    if(!strcmp(weap_ptr->munition->name, weapon))
      return weap_ptr;
    weap_ptr = weap_ptr->next;
  }

  return NULL;
}

/* Get the player to buy n weapon of type `weapon_name', return FALSE if they can't */
int buy_weapon(Player *player, const char *weapon_name, int n)
{
  Weapon *weapon = find_player_weapon(player, weapon_name);
  int cost;

  /* Abort if not a weapon, or the null weapon */
  if(weapon == NULL || weapon == player->weapon_list)
    return FALSE;

  cost = n * weapon->munition->cost;
  if(cost > player->cash)
    return FALSE;
  player->cash -= cost;
  weapon->number += n;

  return TRUE;
}

/* Get the player to sell n weapon of type `weapon_name', return FALSE if they can't */
int sell_weapon(Player *player, const char *weapon_name, int n)
{
  Weapon *weapon = find_player_weapon(player, weapon_name);
  int price;

  /* Abort if not a weapon, or the null weapon */
  if(weapon == NULL || weapon == player->weapon_list)
    return FALSE;

  if(weapon->number < n)
    return FALSE;

  price = 0.5 * n * weapon->munition->cost;
  player->cash += price;
  weapon->number -= n;

  return TRUE;
}

/* Select the weapon `weapon_name' for the player, return FALSE if weapon doesn't exist */
int select_weapon(Player *player, const char *weapon_name)
{
  Weapon *weapon = find_player_weapon(player, weapon_name);

  if(weapon == NULL)
    return FALSE;

  /* If player has some of that weapon, select it */
  if(weapon->number > 0)
    player->selected_weapon = weapon;
  else
    return FALSE;

  return TRUE;
}

/* Move to the next weapon */
void cycle_weapon_fowards(Player *player)
{
  /* If no weapons to cycle through */
  if(player->weapon_list == NULL)
    return;

  /* Cycle fowards until find a weapon that player owns (or the null weapon) */
  do {
    player->selected_weapon = player->selected_weapon->next;
  } while(player->selected_weapon != NULL && player->selected_weapon->number <= 0);

  /* If ran out of weapons, can always select the null weapon */
  if(player->selected_weapon == NULL)
    player->selected_weapon = player->weapon_list;
}

/* Move to the last weapon */
void cycle_weapon_backwards(Player *player)
{
  /* If no weapons to cycle through */
  if(player->weapon_list == NULL)
    return;

  /* Cycle backwards until find a weapon that player owns (or the null weapon) */
  do {
    player->selected_weapon = player->selected_weapon->last;
  } while(player->selected_weapon->number <= 0 &&
	  player->selected_weapon != player->weapon_list);
}

/* Cycle foward n times the players selected weapon. Negative values will cycle backwards */
void cycle_weapon(Player *player, int n)
{
  int i;

  /* If no weapons to cycle through */
  if(player->weapon_list == NULL)
    return;

  /* Cycle fowards */
  for(i = 0; i < n; i++)
    cycle_weapon_fowards(player);
  /* Cycle backwards */
  for(i = 0; i < -n; i++)
    cycle_weapon_backwards(player);
}

/* Clear out all the weapons a player has */
static int clear_weapon_list(Player *player)
{
  Weapon *weapon, *tmp;

  if(!player)
    return FALSE;

  weapon = player->weapon_list;
  while(weapon)
  {
    tmp = weapon->next;
    Nfree(weapon);
    weapon = tmp;
  }

  return TRUE;
}

/* Start the engine */
void start_player_engine(Player *player)
{
  if(player->engine_on)
    return;

  player->drive_time = player->game->driving_time;
  player->has_driven = TRUE;
  player->engine_on = TRUE;
}

/* Stop the engine */
void stop_player_engine(Player *player)
{
  player->engine_on = FALSE;
  set_player_accel(player, 0.0);
}

/* Set the speed of the players tank */
void set_player_speed(Player *player, float speed)
{
  if(!player->engine_on || player->brakes)
    return;

  player->speed = speed;
}

/* Set the speed of the players tank */
void set_player_accel(Player *player, float accel)
{
  float max_acc;

  if(!player->engine_on || player->brakes)
    return;

  max_acc = player->tank->acc;

  player->accel = accel;
  if(player->accel < -max_acc)/*player->tank->reverse_acc)*/
    player->accel = -max_acc;/*player->tank->reverse_acc;*/
  if(player->accel > max_acc)
    player->accel = max_acc;
}

/* Increase the speed of the players tank */
void inc_player_speed(Player *player, float dspeed)
{
  if (player->game->mode != SINGLE_PLAYER)
    if(!player->engine_on || player->brakes)
      return;
  player->speed += dspeed;
}

/* Increase the acceleration of the players tank */
void inc_player_accel(Player *player, float daccel)
{
  float max_accel;

  if(player->game->mode != SINGLE_PLAYER)
    if(!player->engine_on || player->brakes)
      return;

  max_accel = player->tank->acc;

  player->accel += daccel;
  if(player->accel < -max_accel)/*player->tank->reverse_acc)*/
    player->accel = -max_accel;/*player->tank->reverse_acc;*/
  if(player->accel > max_accel)
    player->accel = max_accel;
}

/* Set the brakes for the player */
void set_player_brakes(Player *player, int brakes)
{
  if(!player->engine_on)
    return;

  set_player_accel(player, 0.0);
  player->brakes = brakes;
}

/* Toggle whether the player has the brakes on */
void toggle_player_brakes(Player *player)
{
  if(!player->engine_on)
    return;

  set_player_accel(player, 0.0);
  player->brakes = !player->brakes;
}

/* Increase the heading of the players tank */
void inc_player_heading(Player *player, float dheading)
{
  if(!player->engine_on ||
     !(player->tank_state == ON_GROUND || player->tank_state == ON_SEAFLOOR))
    return;

  player->heading += dheading;
  while(player->heading < -180.0f)
    player->heading += 360.0f;
  while(player->heading > 180.0f)
    player->heading -= 360.0f;

  update_player_direction(player);
}

/* Converts a players rotations to a direction */
static void update_player_direction(Player *player)
{
  GLfloat m[16];

  /* NOTE: This probably isn't the best way to do this -- use proper
     calculations */
  glPushMatrix();
  glLoadIdentity();

  glRotatef(player->rot_angle, player->rot_dir[0], player->rot_dir[1], player->rot_dir[2]);
  glRotatef(player->heading, 0.0, 1.0, 0.0);

  glGetFloatv(GL_MODELVIEW_MATRIX, m);

  player->dir[0] = m[8];
  player->dir[1] = m[9];
  player->dir[2] = m[10];

  glPopMatrix();
}

/* Converts the tanks vel[] to dir[] */
static void update_player_directionv(Player *player)
{
  _vector_dup(player->vel, player->dir);
  normalize_vector(player->dir);
}

/* Set the heading of the players turret */
void set_player_turret_heading(Player *player, float heading)
{
  while(heading > 180.0f)
    heading -= 360.0f;
  while(heading < -180.0f)
    heading += 360.0f;

  player->turret.heading = heading;
}

/* Icrease the heading of the players turret */
void inc_player_turret_heading(Player *player, float dheading)
{
  player->turret.heading += dheading;

  while(player->turret.heading > 180.0f)
    player->turret.heading -= 360.0f;
  while(player->turret.heading < -180.0f)
    player->turret.heading += 360.0f;
}

/* Set the pitch of the players turret */
void set_player_turret_pitch(Player *player, float pitch)
{
  while(pitch < -90.0f)
    pitch += 90.0f;
  while(pitch > 0.0f)
    pitch -= 90.0f;

  player->turret.pitch = pitch;
}

/* Increase the heading of the players turret */
void inc_player_turret_pitch(Player *player, float dpitch)
{
  player->turret.pitch += dpitch;

  if(player->turret.pitch < -90.0f)
  {
    set_player_turret_pitch(player, 180.0f - player->turret.pitch);
    inc_player_turret_heading(player, 180.0f);
  } else if(player->turret.pitch > 0.0f)
    {
      set_player_turret_pitch(player, -1.0f * player->turret.pitch);
      inc_player_turret_heading(player, 180.0f);
    }
}

#ifdef USE_GLASS
/* Set a tanks GLASS object variables to what the player has */
void set_player_tank_variables(Player *player)
{
  Tank *tank = player->tank;
  GlassObject *model = tank->model;

  /* Rotate the tank, and its turret */
  glassSetVariable(model, tank->heading, player->heading);
  glassSetVariable(model, tank->rot, player->rot_angle);
  glassSetVariable(model, tank->rot_x, player->rot_dir[0]);
  glassSetVariable(model, tank->rot_y, player->rot_dir[1]);
  glassSetVariable(model, tank->rot_z, player->rot_dir[2]);
  glassSetVariable(model, tank->turret_pitch, player->turret.pitch);
  glassSetVariable(model, tank->turret_heading, player->turret.heading);
}
#endif

/* Fire the players weapon */
void fire_player(Player *player)
#define MISSILE_INITIAL_SPEED 10.0
{
#ifdef USE_GLASS
  Round *round = player->game->round;
  Tank *tank = player->tank;
  float *barrel_rel_pos, *barrel_dir, barrel_pos[3], *pos = player->pos;
  Munition *munition = player->selected_weapon->munition;
  Projectile *projectile;

  if(!tank->model)
    return;

  /* Update the variables */
  set_player_tank_variables(player);

  /* Get the relative position and direction of the end of the tanks barrel */
  barrel_rel_pos = glassGetActivePointPos(tank->model, tank->barrel);
  barrel_dir = glassGetActivePointDir(tank->model, tank->barrel);

  barrel_pos[0] = barrel_rel_pos[0] + pos[0];
  barrel_pos[1] = barrel_rel_pos[1] + pos[1];
  barrel_pos[2] = barrel_rel_pos[2] + pos[2];

#else

  World *world = player->world;
  Munition *munition = player->selected_weapon->munition;
  Projectile *projectile;
  GLfloat mat[16];
  float barrel_pos[3], barrel_dir[3], vel[3];

  /* Work out vector for end of gun */
  glPushMatrix();
    glLoadIdentity();
    glRotatef(player->rot_angle, player->rot_dir[0], player->rot_dir[1], player->rot_dir[2]);
    glRotatef(player->heading, 0.0, 1.0, 0.0);
    glTranslatef(0.0, 0.0, 0.25);
    glRotatef(player->turret.heading, 0.0, 1.0, 0.0);
    glTranslatef(0.0, 1.75, 0.0);
    glRotatef(player->turret.pitch, 1.0, 0.0, 0.0);
    glTranslatef(0.0, 0.0, 2.5);
    glGetFloatv(GL_MODELVIEW_MATRIX, mat);
  glPopMatrix();

  barrel_pos[0] = mat[12] + player->pos[0];
  barrel_pos[1] = mat[13] + player->pos[1];
  barrel_pos[2] = mat[14] + player->pos[2];
  barrel_dir[0] = mat[8];
  barrel_dir[1] = mat[9];
  barrel_dir[2] = mat[10];
  vel[0] = player->muzzle_vel * barrel_dir[0];
  vel[1] = player->muzzle_vel * barrel_dir[1];
  vel[2] = player->muzzle_vel * barrel_dir[2];
#endif

  /* Set the initial acceleration */
  switch(munition->type)
  {
  case MUNITION_NONE:
    return; /* The null weapon, do nothing */
    break;
  case MUNITION_SHELL:
  case MUNITION_BOMBLET:
    projectile = 
      add_projectile_pdsa(round, munition, barrel_pos, barrel_dir, player->muzzle_vel, 0.0f);
    set_projectile_owner(projectile, player);
    break;
  case MUNITION_MISSILE:
    projectile = add_projectile_pdsa(round,
				     munition,
				     barrel_pos,
				     barrel_dir,
				     MISSILE_INITIAL_SPEED,
				     player->thrust);
    set_projectile_owner(projectile, player);
    break;
  }

  /* Keep track of how much this round cost */
  player->round_cost += munition->cost;

  /* Remove one munition */
  player->selected_weapon->number--;

  /* If that was their last weapon of that type, cycle to next available weapon */
  if(player->selected_weapon->number <= 0)
    cycle_weapon_fowards(player);
}

/* Det the fire direction */
void get_fire_direction(Player *player, float *barrel_dir)
{
#ifdef USE_GLASS
  Tank *tank = player->tank;
  float *glass_barrel_dir;

  if(!tank->model);

  set_player_tank_variables(player);

  glass_barrel_dir = glassGetActivePointDir(tank->model, tank->barrel);

  _vector_dup(glass_barrel_dir, barrel_dir);

#else
  GLfloat mat[16];

  /* Work out vector for end of gun */
  glPushMatrix();
    glLoadIdentity();
    glRotatef(player->rot_angle, player->rot_dir[0], player->rot_dir[1], player->rot_dir[2]);
    glRotatef(player->heading, 0.0, 1.0, 0.0);
    glTranslatef(0.0, 0.0, 0.25);
    glRotatef(player->turret.heading, 0.0, 1.0, 0.0);
    glTranslatef(0.0, 1.75, 0.0);
    glRotatef(player->turret.pitch, 1.0, 0.0, 0.0);
    glTranslatef(0.0, 0.0, 2.5);
    glGetFloatv(GL_MODELVIEW_MATRIX, mat);
  glPopMatrix();

  barrel_dir[0] = mat[8];
  barrel_dir[1] = mat[9];
  barrel_dir[2] = mat[10];
#endif
}

/* Update all the players in a round */
void update_players(Round *round, float dt)
{
  PlayerEntry *entry;
  Player *player;

  if(!round)
    return;

  entry = round->players->first;
  while(entry)
  {
    player = entry->player;
    if(player->player_state == PLAYER_ALIVE)
      update_player(player, dt);

    entry = entry->next;
  }
}

/* Update a player (i.e. move its tank etc) */
static void update_player(Player *player, float dt)
{
  /* Update the players finances */
  update_player_finances(player, dt);

  move_player(player, dt);
}

/* Move the players tank */
static void move_player(Player *player, float dt)
#define TANK_WIDTH 2.0
#define TANK_LENGTH 3.0
#define HALF_TANK_WIDTH (0.5*TANK_WIDTH)
#define HALF_TANK_HEIGHT (0.5*TANK_HEIGHT)
{
  Map *map = player->game->map;
  float *pos = player->pos, *vel = player->vel, *acc = player->acc;
  float *dir = player->dir, vel_dir[3];
  float h0, h1, h2, h3;
  /*float th0, th1, th2, th3;
  int   og0, og1, og2, og3;*/
  /* Vectors for general use */
  float v1[3], v2[3], v3[3];
  float *n;
  float speed, slide, accel;

  /* NOTE: Temporary variables -- from old function */
  float h;

  /*   The players tank    */
  /*                   +x  */
  /* h0 +-----+ h1   +->   */
  /*    |  _  |      |     */
  /*    | |*| |     +y     */
  /*    | \_/ |            */
  /*    |  #  |            */
  /* h3 +--|--+ h2         */
  /*                       */
  /* * = position of tank  */

  /* Get the height of the map at the four
     corners of the tank */
  h0 = get_height(map, pos[0] - HALF_TANK_WIDTH, pos[2] - HALF_TANK_WIDTH);
  h1 = get_height(map, pos[0] + HALF_TANK_WIDTH, pos[2] - HALF_TANK_WIDTH);
  h2 = get_height(map, pos[0] + HALF_TANK_WIDTH, pos[2] + HALF_TANK_WIDTH);
  h3 = get_height(map, pos[0] - HALF_TANK_WIDTH, pos[2] + HALF_TANK_WIDTH);

  /* Get the height of those corners on the tank */
  /* ... */

  /* Update the position */
  pos[0] += dt * (vel[0] + 0.5f * acc[0] * dt);
  pos[1] += dt * (vel[1] + 0.5f * acc[1] * dt);
  pos[2] += dt * (vel[2] + 0.5f * acc[2] * dt);

  /* Bounds checking */
  /* Note: never sit exactly on the edge as this will stuff get_height() when converting to
     an int */
  if(pos[0] < 0.0f)
  {
    pos[0] = 0.0f;
    vel[0] = 0.0f;
  }
  if(pos[0] >= map->width)
  {
    pos[0] = map->width - 0.001;
    vel[0] = 0.0f;
  }
  if(pos[2] < 0.0f)
  {
    pos[2] = 0.0f;
    vel[2] = 0.0f;
  }
  if(pos[2] >= map->height)
  {
    pos[2] = map->height - 0.001;
    vel[2] = 0.0f;
  }

  /* Check if the player is above or below ground/underwater */
  h = get_height(map, pos[0], pos[2]);
  if(pos[1] <= h)
  {
    pos[1] = h;
    vel[1] = 0.0f;

    player->tank_state = ON_GROUND;
  }
  else
  {
    if((vel[1] < -10.0f))
      player->tank_state = PARACHUTE;
    else
      if((vel[1] >= 0.0f) || (player->tank_state != PARACHUTE))
	player->tank_state = IN_AIR;
  }

  if(pos[1] < player->game->effective_sea_level)
  {
    if(player->tank_state == ON_GROUND)
      player->tank_state = ON_SEAFLOOR;
    else
      player->tank_state = UNDERWATER;
  }
  else
    player->underwater_time = 0.0;

  /* Check the player hasn't run out of driving time */
  if(player->game->mode == TURN_BASED)
  {
    if(player->engine_on)
    {
      player->drive_time -= dt;
      if(player->drive_time < 0.0)
      {
        /* Put the brakes on */
        set_player_brakes(player, TRUE);
        /* Times up! */
        player_end_turn(player);
      }
    }
  }
  else if (player->game->mode == SINGLE_PLAYER)
  {/* in single player we can always move */
  }

  /* Base acceleration -- gravity */
  acc[0] = 0.0;
  acc[1] = player->game->gravity;
  acc[2] = 0.0;

  switch(player->tank_state)
  {
  case ON_GROUND:
  case ON_SEAFLOOR:
    /* The direction the tank is moving in */
    _vector_dup(vel, vel_dir);
    normalize_vector(vel_dir);

    /* The speed of the tank in the direction it's pointing */
    speed = vector_dot(dir, vel);

    /* Get the velocity vector in the direction the tank is pointing */
    _vector_scale(speed, dir, v1);
    /* Get the velocity vector orthogonal to the direction of the tank */
    _vector_sub(vel, v1, v2);

    /* The speed of the slide */
    slide = vector_2norm(v2);

    /* If going too fast, slow down */
    if(speed > player->tank->speed)
      accel = 0.1 * dt * (player->tank->speed - speed);
    else
      accel = player->accel;

    /* Accelerate the tank in the direction it's pointing */
    _vector_scale(accel, dir, v3);
    _vector_add(acc, v3, acc);

    /* Slow the tank when sliding sideways */
    if(slide < 0.1) /* Stop the slide completely */
      _vector_sub(vel, v2, vel);
    else /* Slow the slide down */
    {
      /* Simplified, as would be:
	 (-k * slide * slide) scaling the unit slide vector,
	 but v2 is (slide*unit slide), so scale by 1/slide,
	 giving below */
      _vector_scale(-0.5*slide, v2, v3);
      _vector_add(acc, v3, acc);
    }

    /* Apply the brakes */
    if(player->brakes)
    {
      /* Stopped */
      if(fabs(speed) < 0.1)
      {
	vel[0] = 0.0;
	vel[1] = 0.0;
	vel[2] = 0.0;
	acc[0] = 0.0;
	acc[1] = player->game->gravity;
	acc[2] = 0.0;
	set_player_brakes(player, FALSE);
      }
      else /* Slow down */
      {
	/* Coefficent of drag is proportional to k/fabs(speed),
	   hence this calculation is simplified
	   would be k/fabs(speed) * speed*fabs(speed) */
	_vector_scale(-4.0 * speed, dir, v3);
	_vector_add(acc, v3, acc);
      }
    }

    if(player->tank_state == ON_SEAFLOOR)
    {
      /* Extra drag from being underwater */
      speed = vel[0]*vel[0] + vel[1]*vel[1] + vel[2]*vel[2];
      _vector_scale(-0.5 * speed, vel_dir, v3);
      _vector_add(acc, v3, acc);

      player->underwater_time += dt;
      /* if under water since a too long time, reduce health */
      if(player->underwater_time > player->tank->underwater_ability)
	player->health -= player->tank->underwater_damage*dt;
    }

    n = get_smooth_normal(map, player->pos[0], player->pos[2]);

    /* Take the cross-product of the tanks up vector, and the map normal,
       to give the direction to rotate around.
       Since up = (0,1,0), this simplifys to: */
    _make_vector(n[2], 0.0, -n[0], player->rot_dir);

    /* The angle to rotate is acos(n dot up), i.e. (simplified) */
    player->rot_angle = (180.0 / M_PI) * acos(n[1]);

    update_player_direction(player);
    break;
  case IN_AIR:
    break;
  case PARACHUTE:
    /* Only decrease vertical motion -- parapenting <sp> */
    player->acc[1] += -0.3 * player->vel[1] * fabs(player->vel[1]);
    break;
  case UNDERWATER:
    /* The direction the tank is moving in */
    _vector_dup(vel, vel_dir);
    normalize_vector(vel_dir);

    player->underwater_time += dt;
    /* if under water since a too long time, reduce health */
    if(player->underwater_time > player->tank->underwater_ability)
      player->health -= player->tank->underwater_damage*dt;

    /* Drag the tank as it falls underwater */
    /* The speed of the tank in the direction it's pointing */
    speed = vel[0]*vel[0] + vel[1]*vel[1] + vel[2]*vel[2];
    _vector_scale(-0.5 * speed, vel_dir, v3);
    _vector_add(acc, v3, acc);
    break;
  default:
    Neprintf("Error: Player in unknown state\n");
    break;
  }

  /* Update velocities */
  player->vel[0] += player->acc[0] * dt;
  player->vel[1] += player->acc[1] * dt;
  player->vel[2] += player->acc[2] * dt;
}

/* Update the players finances */
static void update_player_finances(Player *player, float dt)
#define TOTAL_DISPLAY_TIME 8.0
#define FADE_IN_TIME 2.0
#define FADE_OUT_TIME 2.0
#define COUNT_TIME (TOTAL_DISPLAY_TIME - FADE_IN_TIME - FADE_OUT_TIME)
{
  float timer;

  /* Update the players winnings */
  if(player->cash_display_timer && player == player->game->round->current_player)
  {
    player->cash_display_timer += dt;
    timer = player->cash_display_timer;

    /* Stay visible for TOTAL_DISPLAY_TIME secs */
    if(timer > TOTAL_DISPLAY_TIME)
      player->cash_display_timer = 0.0;
    /* Convert all winnings to cash in before fades out */
    else if(timer > (TOTAL_DISPLAY_TIME - FADE_OUT_TIME))
    {
      player->cash += player->winnings;
      player->winnings = 0;

      player->percent_transferred = 0.0;

      player->temp_cash = player->cash;
      player->temp_winnings = 0;
    }
    else if(timer > FADE_IN_TIME)
    {
      player->percent_transferred +=
	dt * (1.0 - player->percent_transferred) / (TOTAL_DISPLAY_TIME - FADE_OUT_TIME - timer);
      player->temp_winnings =
	(int)((1.0 - player->percent_transferred) * (float)player->winnings);
      player->temp_cash = player->cash + player->winnings - player->temp_winnings;
    }
  }
}

/* Store permamently all a players floating cash */
int bank_cash(Player *player)
{
  if(!player)
    return FALSE;

  player->cash += player->winnings;
  player->temp_cash = player->cash;
  player->winnings = 0;
  player->temp_winnings = 0;
  player->cash_display_timer = 0.0;

  return TRUE;
}

/* Draw all the players in a round */
void draw_players(Round *round)
{
  GLfloat white[4] = {1.0, 1.0, 1.0, 1.0}, black[4] = {0.0, 0.0, 0.0, 1.0};
  GLfloat c[4] = {0.6, 0.5, 0.3, 1.0};
  Player *player;
  PlayerEntry *pentry;

  if(round->world->is_texture)
  {
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, white);
    glColor3fv(white);
  }
  else
  {
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, c);
    glColor3fv(c);
  }

  pentry = round->players->first;
  while(pentry)
  {
    player = pentry->player;

    if(player->player_state == PLAYER_ALIVE)
      draw_player(player);

    pentry = pentry->next;
  }

  /* Set the emission, and specular to defaults */
  glMaterialfv(GL_FRONT, GL_EMISSION, black);
  glMaterialfv(GL_FRONT, GL_SPECULAR, black);
}

/* Draw the player */
void draw_player(Player *player)
{
#ifdef USE_GLASS
  Tank *tank = player->tank;
  GlassObject *model = tank->model;
  float *pos = player->pos;

  draw_ai(player->ai);

  if(!model)
    return;

  glPushMatrix();

  /* Translate to where the tank is on the map */
  glTranslatef(pos[0], pos[1], pos[2]);

  /* Parachute */
  if(player->tank_state == PARACHUTE)
  {
    glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT);

    glDisable(GL_TEXTURE_2D);

    glPushMatrix();
    glTranslatef(0.0f, 1.75f, 0.0f);
    glRotatef(90.0f, 1.0f, 0.0f, 0.0f);
    glScalef(15.0f, 15.0f, 15.0f);
    glCallList(player->world->lists.parachute);
    glPopMatrix();

    if(player->world->is_texture)
      glEnable(GL_TEXTURE_2D);

    glPopAttrib();
  }

  /* Set the variables */
  set_player_tank_variables(player);

  /* Draw the tank */
  glassDrawObject(model);

  glPopMatrix();

#else
  DisplayLists *lists = &player->world->lists;

  glPushMatrix();

  glTranslatef(player->pos[0], player->pos[1], player->pos[2]);

  /* Rotate player */
  glRotatef(player->rot_angle, player->rot_dir[0], player->rot_dir[1], player->rot_dir[2]);
  glRotatef(player->heading, 0.0f, 1.0f, 0.0f);

  /* Parachute */
  if(player->tank_state == PARACHUTE)
  {
    glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT);

    glDisable(GL_TEXTURE_2D);

    glPushMatrix();
    glTranslatef(0.0f, 1.75f, 0.0f);
    glRotatef(90.0f, 1.0f, 0.0f, 0.0f);
    glScalef(15.0f, 15.0f, 15.0f);
    glCallList(lists->parachute);
    glPopMatrix();

    if(player->world->is_texture)
      glEnable(GL_TEXTURE_2D);

    glPopAttrib();
  }

  glBindTexture(GL_TEXTURE_2D, player->tank->texture);

  /* Body */
  glCallList(lists->body);

  /* Turret */
  glTranslatef(0.0f, 0.0f, 0.25);
  glRotatef(player->turret.heading, 0.0f, 1.0f, 0.0f);
  glCallList(lists->turret);

  /* Gun */
  glTranslatef(0.0f, 1.75, 0.0f);
  glRotatef(player->turret.pitch, 1.0f, 0.0f, 0.0f);
  glCallList(lists->gun);

  glPopMatrix();
#endif
}

/* Remove all the players from the world */
int remove_players(World *world)
{
  Player *player = world->player_list, *tmp_player;

  while(player != NULL)
  {
    tmp_player = player->next;
    remove_player(player);
    player = tmp_player;
  }
  world->player_list = NULL;

  return TRUE;
}

/* Remove a player from the world */
void remove_player(Player *player)
{
  if(player->next != NULL)
    player->next->last = player->last;
  if(player->last != NULL)
    player->last->next = player->next;
  else
    player->world->player_list = NULL;

  Nprintf("Player \"%s\" removed\n", player->name);

  free_player(player);
}

/* Free a player */
static void free_player(Player *player)
{
  Nfree(player->name);

  free_ai(player->ai);
  clear_weapon_list(player);
  free_player_stats(player);

  Nfree(player);
}
