/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __player_h__
#define __player_h__

typedef enum   _PlayerState PlayerState;
typedef enum   _PlayerType PlayerType;
typedef enum   _TankState TankState;
typedef struct _Turret Turret;
typedef struct _Player Player;
typedef struct _Weapon Weapon;

#include "world.h"
#include "player_stats.h"
#include "ai.h"
#include "tank.h"
#include "munition.h"
#include "round.h"

enum _PlayerState
{
  PLAYER_DEAD = 0,
  PLAYER_ALIVE
};

enum _PlayerType
{
  PLAYER_HUMAN = 0,
  PLAYER_AI_STUPID,
  PLAYER_AI_MEDIUM,
  PLAYER_AI_GOOD,
  PLAYER_AI_GOD,
  /* Keep this at the end, so is not selectable in
     player config screen */
  PLAYER_HUMAN_NET
};

enum _TankState
{
  ON_GROUND = 0,
  IN_AIR,
  PARACHUTE,
  UNDERWATER,
  ON_SEAFLOOR
};

struct _Turret
{
  float heading, pitch;
};

struct _Player
{
  World *world;
  Game *game;

  char *name;

  /* How much cash they have, and how much they've won,
     and how much of those winnings has been added to the
     total cash */
  int cash, winnings, temp_cash, temp_winnings;
  float percent_transferred, cash_display_timer;

  /* How much health they have */
  float health;

  /* How much this round cost them, and how much they won */
  int round_cost, round_winnings;

  /* How many other players they've killed this round, and
     how many this game */
  int round_kills, kills;

  /* The statistics on the other players */
  PlayerStat *player_stats;

  /* If they're dead/alive/not playing etc */
  PlayerState player_state;

  /* If they're human, and AI, or network player */
  PlayerType player_type;

  /* AI internal state */
  AI *ai;

  /* The availiable weapons to use */
  Weapon *weapon_list;
  Weapon *selected_weapon;

  /* What they're driving */
  Tank *tank;
  /* Whether the tank is in the air/using a parachute/on the ground/underwater */
  TankState tank_state;

  /* The engine settings */
  int engine_on, has_driven;
  float drive_time;
  int brakes;
  float speed, accel;
  /* The current position/velocity/acceleration */
  float pos[3], vel[3], acc[3];

  /* The direction the tank is pointing in, as a vector */
  float dir[3];

  /* The heading of the tank, and the rotation around a
     vector to keep tank flat on ground */
  float heading, rot_dir[3], rot_angle;

  /* Where the turret (gun) is pointing */
  Turret turret;
  /* How fast they're firing their next weapon */
  float muzzle_vel, thrust;

  float underwater_time;

  Player *last, *next;
};

/* The amount of a type of munition the player has */
struct _Weapon
{
  Munition *munition;
  int number;

  Weapon *next, *last;
};

/* Add a new player into the world */
Player *add_player(World *world, const char *name);

/* Set the players name */
void set_player_name(Player *player, const char *name);

/* Set the amount of cash the player has */
void set_player_cash(Player *player, int cash);

/* Give the player some winnings */
void give_player_winnings(Player *player, int winnings);

/* Set the type of the player (human, ai...) */
void set_player_type(Player *player, PlayerType type);

/* Set the players tank, return FALSE if tank not found */
int set_player_tank(Player *player, const char *unit_name);

/* Get the player to buy n weapon of type `weapon_name', return FALSE if they can't */
int buy_weapon(Player *player, const char *weapon_name, int n);

/* Get the player to sell n weapon of type `weapon_name', return FALSE if they can't */
int sell_weapon(Player *player, const char *weapon_name, int n);

/* Select the weapon `weapon_name' for the player, return FALSE if weapon doesn't exist */
int select_weapon(Player *player, const char *weapon_name);

/* Move to the next weapon */
void cycle_weapon_fowards(Player *player);

/* Move to the last weapon */
void cycle_weapon_backwards(Player *player);

/* Cycle foward n times the players selected weapon. Negative values will cycle backwards */
void cycle_weapon(Player *player, int n);

/* Start the engine */
void start_player_engine(Player *player);

/* Stop the engine */
void stop_player_engine(Player *player);

/* Set the speed of the players tank */
void set_player_speed(Player *player, float speed);

/* Set the speed of the players tank */
void set_player_accel(Player *player, float accel);

/* Increase the speed of the players tank */
void inc_player_speed(Player *player, float dspeed);

/* Increase the acceleration of the players tank */
void inc_player_accel(Player *player, float daccel);

/* Set the brakes for the player */
void set_player_brakes(Player *player, int brakes);

/* Toggle whether the player has the brakes on */
void toggle_player_brakes(Player *player);

/* Increase the heading of the players tank */
void inc_player_heading(Player *player, float dheading);

/* Set the heading of the players turret */
void set_player_turret_heading(Player *player, float heading);

/* Icrease the heading of the players turret */
void inc_player_turret_heading(Player *player, float dheading);

/* Set the pitch of the players turret */
void set_player_turret_pitch(Player *player, float pitch);

/* Increase the heading of the players turret */
void inc_player_turret_pitch(Player *player, float dpitch);

#ifdef USE_GLASS
/* Set a tanks GLASS object variables to what the player has */
void set_player_tank_variables(Player *player);
#endif

/* Fire the players weapon */
void fire_player(Player *player);

/* get the fire direction (used by AI) */
void get_fire_direction(Player *player, float *barrel_dir);

/* Update all the players in a round */
void update_players(Round *round, float dt);

/* Store permamently all a players floating cash */
int bank_cash(Player *player);

/* Draw all the players in the world */
void draw_players(Round *round);

/* Draw the player */
void draw_player(Player *player);

/* Remove all the players from the world */
int remove_players(World *world);

/* Remove a player from the world */
void remove_player(Player *player);

#endif /* __player_h__ */
