/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* projectile.c
   This file contains the functions to create and
   manage projectiles */

#include <stdio.h>
#include <math.h>
#include <malloc.h>

#include "projectile.h"
#include "world.h"
#include "game.h"
#include "vector.h"
#include "map.h"
#include "trails.h"
#include "munition.h"
#include "explosion.h"
#include "text.h"
#include "tools.h"
#include "light.h"

static void update_projectile(Projectile *, float);
static void move_projectile(Projectile *, float);
static void calculate_projectile_angles(Projectile *);
static void calculate_projectile_direction(Projectile *);
static int check_projectile_collision(Projectile *projectile);
static int projectile_event(Projectile *, MunitionAction *);
static void draw_projectile(Projectile *);
static int free_projectile(Projectile *projectile);

/* *********************************************** */
/*            PROJECTILE ADD/REMOVE CODE           */
/* *********************************************** */

/* Add a projectile based on three vectors, position (displacement), velocity (momentum), and
   acceleration (force) */
Projectile *add_projectile_dmf(Round *round,
			       Munition *munition,
			       float *pos,
			       float *vel,
			       float *acc)
{
  /* !!note acc[] is not currently used */
  return add_projectile(round, munition, pos, vel, NULL, 0.0f);
}

/* Add a projectile based on two vectors, position and direction, and two scalars: linear speed
   and acceleration */
Projectile *add_projectile_pdsa(Round *round,
				Munition *munition,
				float *pos,
				float *dir,
				float spd,
				float thrust)
{
  float vel[3];

  vel[0] = spd * dir[0];
  vel[1] = spd * dir[1];
  vel[2] = spd * dir[2];

  return add_projectile(round, munition, pos, vel, dir, thrust);
}

/* Add a projectile with all it's requirements, type, pos[], vel[], dir[], and acc
   Either the direction or the velocity can be sent as NULL, they can be created from
   eachother */
Projectile *add_projectile(Round *round,
			   Munition *munition,
			   float *pos,
			   float *vel,
			   float *dir,
			   float thrust)
{
  Projectile *projectile;
  int i;

  projectile = malloc(sizeof(Projectile));
  if(!projectile)
  {
    Neprintf("Error: Unable to allocate space for projectile");
    return NULL;
  }

  /* Link the projectile into the correct world */
  projectile->world = round->game->world;
  projectile->game = round->game;
  projectile->round = round;

  /* Owned by no-one */
  projectile->owner = NULL;

  /* Launch without a parachute */
  projectile->parachute_deployed = FALSE;

  /* Set the thrust and drag */
  projectile->thrust = thrust;
  projectile->drag = 0.0;
  projectile->parachute_drag = 0.0;

  /* Set the default position */
  projectile->pos = projectile->p0;
  projectile->old_pos = projectile->p1;
  projectile->collision_pos = projectile->p2;
  projectile->in_collision = FALSE;
  for(i = 0; i < 3; i++)
  {
    projectile->p0[i] = pos[i];
    projectile->p1[i] = pos[i];
    projectile->p2[i] = 0.0;
    projectile->rot[i] = 0.0;
  }

  /* Set the initial velocity */
  if(vel)
  {
    projectile->vel[0] = vel[0];
    projectile->vel[1] = vel[1];
    projectile->vel[2] = vel[2];
  }
  else
  {
    projectile->vel[0] = 0.0;
    projectile->vel[1] = 0.0;
    projectile->vel[2] = 0.0;
  }

  /* Set the initial acceleration */
  /* NOTE: Not currently used */
  projectile->acc[0] = 0.0f;
  projectile->acc[1] = 0.0f;
  projectile->acc[2] = 0.0f;

  /* Set the initial direction */
  if(dir)
  {
    projectile->dir[0] = dir[0];
    projectile->dir[1] = dir[1];
    projectile->dir[2] = dir[2];
  } else
    calculate_projectile_direction(projectile);

  /* Get an initial up vector */
  projectile->up[0] = 0.0;
  projectile->up[1] = -projectile->dir[2];
  projectile->up[2] = projectile->dir[1];

  /* The last trail properties */
  projectile->old_trail_pos[0] = pos[0];
  projectile->old_trail_pos[1] = pos[1];
  projectile->old_trail_pos[2] = pos[2];
  projectile->old_trail_up[0] = projectile->up[0];
  projectile->old_trail_up[1] = projectile->up[1];
  projectile->old_trail_up[2] = projectile->up[2];
  _vector_cross(projectile->up, projectile->dir, projectile->old_trail_left);

  /* Reset the trail timer */
  /* TODO: Make this variable */
  projectile->trail_time = TRAIL_TIME;

  /* Calculate the initial rotation */
  calculate_projectile_angles(projectile);

  /* Reset the timers */
  projectile->old_timer = -1.0f;
  projectile->timer = 0.0f;
  projectile->parachute_timer = 0.0f;

  /* Set the projectiles mottion state */
  projectile->state = RISING;

  /* Attach projectile to the start of the list (so they can be spawned in the physics
     function and not be moved) */
  /* Link the projectiles */
  projectile->next = round->projectile_list;
  projectile->last = NULL;

  /* Make the projectile the correct munition */
  projectile->munition = munition;

  /* Point the old first projectile back at the new one */
  if(round->projectile_list != NULL)
    round->projectile_list->last = projectile;
  /* Stick it to the start of the list */
  round->projectile_list = projectile;

  Nprintf("Projectile added (%s)\n", munition->name);

  /* Update max_dt for the current weapons */
  /* FIXME */
  /*update_max_dt(world);*/

  return projectile;
}

/* Set the owner of the projectile */
void set_projectile_owner(Projectile *projectile, Player *owner)
{
  projectile->owner = owner;
}

/* Bye-bye projectile */
void remove_projectile(Projectile *projectile)
{
  Round *round = projectile->round;

  /* Tie the projectile on either side together */
  if(projectile->last == NULL) /* if the first projectile in the list */
    round->projectile_list = projectile->next;
  else
    projectile->last->next = projectile->next;
  if(projectile->next != NULL) /* if not the last projectile in the list */
    projectile->next->last = projectile->last;

  Nprintf("Projectile removed. (%s)\n", projectile->munition->name);

  /* Remove the memory used by the projectile */
  free_projectile(projectile);

  /* Update max_dt for the current weapons */
  /* FIXME */
  /*update_max_dt(world);*/
}

/* Slows down time if weapons that need a minimum accuracy are present */
void update_max_dt(Round *round)
{
  Projectile *proj_ptr = round->projectile_list;

  /* FIXME */
  /*world->max_dt = 10000.0;*/

  while(proj_ptr != NULL)
  {
    /* Not too fast with parachutes around -- otherwise the
       feedback will be too great -> overflow errors */
    /* FIXME */
    /*if(proj_ptr->parachute_deployed)
      world->max_dt = 0.2;*/

    proj_ptr = proj_ptr->next;
  }
}

/* ******************************************* */
/*            PROJECTILE UPDATE CODE           */
/* ******************************************* */

/* Updates all the projectiles in the world */
void update_projectiles(Round *round, float dt)
{
  Projectile *proj_ptr = round->projectile_list, *temp_proj;

  /* Go through projectiles */
  while(proj_ptr != NULL)
  {
    /* Projectile may be removed during move_projectile() so get the next one now */
    temp_proj = proj_ptr->next;
    update_projectile(proj_ptr, dt);
    proj_ptr = temp_proj;
  }

  return;
}

/* Updates (moves, tests for events, does events etc) a projectile */
static void update_projectile(Projectile *projectile, float dt)
{
  float critical_time;

  /* Increase projectile timer */
  projectile->old_timer = projectile->timer;
  projectile->timer += dt;

  /* Decrease time left for next trail */
  projectile->trail_time -= dt;

  /* Move the projectile */
  move_projectile(projectile, dt);

  /* Test for a collision event */
  if(check_projectile_collision(projectile))
    if(projectile_event(projectile, projectile->munition->collision_actions))
      return;

  /* To prevent weapons flying for ever */
  /* Note: This could be put in the munitions file, but if someone forgot it,
     it would be a pain */
  if(projectile->timer > 20.0)
  {
    /* life: 20 sec. */
    Nprintf("Removing projectile (too old)\n");
    remove_projectile(projectile);
    return;
  }

  /* Test for a time event */
  critical_time = projectile->munition->critical_time;
  if((projectile->timer > critical_time) && (projectile->old_timer < critical_time))
    if(projectile_event(projectile, projectile->munition->timer_actions))
      return;

  /* Test for a apex event */
  if(projectile->vel[1] < 0.0f)
  {
    if(projectile->state == RISING)
    {
      if(projectile_event(projectile, projectile->munition->apex_actions))
	return;
      projectile->state = FALLING;
    }
  }
}

/* Move an individual projectile */
static void move_projectile(Projectile *projectile, float dt)
/* Time taken to deploy parachute 33% */
#define PARACHUTE_DEPLOY_TIME 1.0
{
  float *pos, *old_pos, *vel, *acc, *dir, *rot, *drag, thrust;
  float *tpos;

  /* Swap the old position and current position arrays */
  /* This gives the correct old position, and then the
     new position is updated */
  tpos = projectile->old_pos;
  projectile->old_pos = projectile->pos;
  projectile->pos = tpos;

  pos = projectile->pos;
  old_pos = projectile->old_pos;
  vel = projectile->vel;
  acc = projectile->acc;
  dir = projectile->dir;
  rot = projectile->rot;
  drag = &projectile->drag;
  thrust = projectile->thrust;

  /* Update the acceleration */
  acc[0] = dir[0] * thrust;
  acc[1] = (dir[1] * thrust) + projectile->game->gravity;
  acc[2] = dir[2] * thrust;

  /* If parachute is deploying, increase the parachute drag */
  if(projectile->parachute_deployed && *drag != projectile->parachute_drag)
  {
    projectile->parachute_timer += dt;
    /* !!note: deploys in PARACHUTE_DEPLOY_TIME, make variable */
    *drag = (1.0 - exp(-projectile->parachute_timer / PARACHUTE_DEPLOY_TIME))
      * projectile->parachute_drag;
  }
  else
    *drag = 0.0;

  /* If drag is present, calculate its effect */
  /* NOTE: only drag from parachutes at the moment, make two+ types */
  if(*drag)
  {
    acc[0] -= 0.5 * *drag * vel[0] * fabs(vel[0]);
    acc[1] -= 0.5 * *drag * vel[1] * fabs(vel[1]);
    acc[2] -= 0.5 * *drag * vel[2] * fabs(vel[2]);
  }

  /* Update the position */
  pos[0] = old_pos[0] + dt * (vel[0] + 0.5*acc[0]*dt);
  pos[1] = old_pos[1] + dt * (vel[1] + 0.5*acc[1]*dt);
  pos[2] = old_pos[2] + dt * (vel[2] + 0.5*acc[2]*dt);

  /* Update the velocity */
  vel[0] += acc[0] * dt;
  vel[1] += acc[1] * dt;
  vel[2] += acc[2] * dt;

  /* Update the direction */
  calculate_projectile_direction(projectile);

  /* Update the rotation */
  rot[2] += 360.0f * dt;
  if(rot[1] > 360.0f)
    rot[2] -= 360.0f;
  calculate_projectile_angles(projectile);
}

/* Calculate the heading and pitch of the projectile based on its velocity */
static void calculate_projectile_angles(Projectile *projectile)
{
  float *vel = projectile->vel, *rot = projectile->rot;

  /* Calculate the heading */
  if(vel[2] != 0.0f) /* This doesn't appear to be necessary in practise, but here just in case */
    rot[0] = (180.0f / M_PI) * atan(vel[0] / vel[2]);
  else rot[0] = 0.0f; /* !!note: wrong (should be +-90) */
  if(vel[2] < 0.0f)
  {
    if(vel[0] > 0.0f)
      rot[0] += 180.0f;
    else
      rot[0] -= 180.0f;
  }
    
  /* Calculate the pitch */
  if(!(vel[0] == 0.0f && vel[2] == 0.0f))
    rot[1] = -(180.0f / M_PI) * atan(vel[1] / sqrt(vel[0]*vel[0] + vel[2]*vel[2]));
  else
    rot[1] = 0.0f;
}

/* Make a unit direction vector out of the projectiles velocity */
static void calculate_projectile_direction(Projectile *projectile)
{
  _vector_dup(projectile->vel, projectile->dir);
  normalize_vector(projectile->dir);
}

/* Tests for a collision and places the smoke trails */
static int check_projectile_collision(Projectile *projectile)
{
  Round *round = projectile->round;
  Map *map = projectile->game->map;
  float *p, left[3];

  /* Get the collision point */
  p = get_map_intersection(map, projectile->old_pos, projectile->pos, -1.0);

  if(p) /* If there was a collision */
  {
    /* Get the current left vector for the projectile */
    _vector_cross(projectile->up, projectile->dir, left);

    /* Add the smoke trail to the collision point */
    add_trail(round,
	      projectile->old_trail_pos, p,
	      projectile->old_trail_up, projectile->up,
	      projectile->old_trail_left, left,
	      NULL, 0.2, TRAIL_TIME);

    /* Store the old position, up, and left vectors */
    _vector_dup(projectile->pos, projectile->old_trail_pos);
    _vector_dup(projectile->up, projectile->old_trail_up);
    _vector_dup(left, projectile->old_trail_left);

    /* Store the collision point */
    _vector_dup(p, projectile->collision_pos);
    projectile->in_collision = TRUE;

    return TRUE;
  }
  else /* A miss */
  {
    if(projectile->trail_time < 0.0)
    {
      /* Get the current left vector for the projectile */
      _vector_cross(projectile->up, projectile->dir, left);

      /* Add a piece of a smoke trail */
      add_trail(round,
		projectile->old_trail_pos, projectile->pos,
		projectile->old_trail_up, projectile->up,
		projectile->old_trail_left, left,
		NULL, 0.2, TRAIL_TIME);

      /* Store the old position, up, and left vectors */
      _vector_dup(projectile->pos, projectile->old_trail_pos);
      _vector_dup(projectile->up, projectile->old_trail_up);
      _vector_dup(left, projectile->old_trail_left);

      /* Reset the timer */
      /* TODO: Make this variable */
      projectile->trail_time = TRAIL_TIME;
    }

    /* If out of bounds, and underground, then there _is_ a collision */
    if(projectile->pos[1] < 0.0f && !in_boundsv(map, projectile->pos))
    {
      /* Store the collision point */
      _vector_dup(projectile->pos, projectile->collision_pos);
      projectile->in_collision = TRUE;

      return TRUE;
    }
    return FALSE;
  }

  /* Can't get here */
  return FALSE;
}

/* Cause a projectile to execute an action list, return TRUE if projectile is removed */
static int projectile_event(Projectile *projectile, MunitionAction *action_list)
{
  Round *round = projectile->round;
  Map *map = projectile->game->map;
  MunitionAction *action_ptr = action_list;
  Munition *munition = NULL;
  Projectile *spawn_proj;
  int i;
  float *t, *n, scale, dv[3], v[3], d[3];

  /* Go through the different actions */
  while(action_ptr != NULL)
  {
    switch(action_ptr->type)
    {
    case MUNITION_UNKNOWN_ACTION:
      Neprintf("Warning: Unknown action tried\n");
      break;
    case MUNITION_DELETE:
      /* !!note: temp - Clear out the light */
      clear_dynamic_light(map, projectile->old_pos, 100.0);
      /* Remove the projectile */
      Nprintf("Removing projectile (self deleted)\n");
      remove_projectile(projectile);
      /* Projectile no longer exists, so return TRUE (is dead) */
      return TRUE;
      break;
    case MUNITION_BOUNCE:
      /* The normal at the collision point */
      n = get_normal(map, projectile->pos[0], projectile->pos[2]);
      /* The amount of speed in the direction of the normal */
      scale = -vector_dot(projectile->vel, n);
      /* The required change in velocity to point a percentage
	 of the initial velocity backwards */
      _vector_scale((1.0 + action_ptr->parameters[0].float_param) * scale, n, dv);

      /* Reflect the velocity off the normal */
      projectile->vel[0] += dv[0];
      projectile->vel[1] += dv[1];
      projectile->vel[2] += dv[2];

      /* Velocity has changed so recalculate direction */
      calculate_projectile_direction(projectile);

      break;
    case MUNITION_SPAWN_NEW:
      /* Make a new projectile in the same position with the same owner */
      spawn_proj = add_projectile_dmf(round,
				      find_munition(projectile->munition,
						    action_ptr->parameters[0].string_param),
				      projectile->pos, projectile->vel, NULL);
      set_projectile_owner(spawn_proj, projectile->owner);
      break;
    case MUNITION_SPAWN_NEW_MULTIPLE:
      for(i = 0; i < action_ptr->parameters[1].int_param; i++)
      {
	spawn_proj = add_projectile_dmf(round,
					find_munition(projectile->munition,
						      action_ptr->parameters[0].string_param),
					projectile->pos, projectile->vel, NULL);
	set_projectile_owner(spawn_proj, projectile->owner);
      }
      break;
    case MUNITION_SPAWN_NEW_RANDOM:
      _make_random_vector(d);
      _vector_scale(action_ptr->parameters[1].float_param, d, dv);
      _vector_add(projectile->vel, dv, v);

      spawn_proj = add_projectile_dmf(round,
				      find_munition(projectile->munition,
						    action_ptr->parameters[0].string_param),
				      projectile->pos, v, NULL);
      set_projectile_owner(spawn_proj, projectile->owner);
      break;
    case MUNITION_SPAWN_NEW_RANDOM_MULTIPLE:
      munition = find_munition(projectile->munition,
			       action_ptr->parameters[0].string_param);
      for(i = 0; i < action_ptr->parameters[1].int_param; i++)
      {
	_make_random_vector(d);
	_vector_scale(action_ptr->parameters[2].float_param, d, dv);
	_vector_add(projectile->vel, dv, v);

	spawn_proj = add_projectile_dmf(round, munition, projectile->pos, v, NULL);
	set_projectile_owner(spawn_proj, projectile->owner);
      }
      break;
    case MUNITION_SPAWN_NEW_CLONE_RANDOM_MULTIPLE:
      for(i = 0; i < action_ptr->parameters[1].int_param; i++)
      {
	_make_random_vector(d);
	_vector_scale(action_ptr->parameters[1].float_param, d, dv);
	_vector_add(projectile->vel, dv, v);

	spawn_proj = add_projectile_dmf(round, projectile->munition, projectile->pos, v, NULL);
	set_projectile_owner(spawn_proj, projectile->owner);
      }
    case MUNITION_SPAWN_NEW_CLONE:
      spawn_proj = add_projectile_dmf(round,
				      projectile->munition,
				      projectile->pos,
				      projectile->vel, NULL);
      set_projectile_owner(spawn_proj, projectile->owner);
      break;
    case MUNITION_SPAWN_NEW_CLONE_MULTIPLE:
      for(i = 0; i < action_ptr->parameters[0].int_param; i++)
      {
	spawn_proj = add_projectile_dmf(round,
					projectile->munition,
					projectile->pos,
					projectile->vel, NULL);
	set_projectile_owner(spawn_proj, projectile->owner);
      }
      break;
    case MUNITION_SPAWN_NEW_CLONE_RANDOM:
      _make_random_vector(d);
      _vector_scale(action_ptr->parameters[0].float_param, d, dv);
      _vector_add(projectile->vel, dv, v);

      /* Make a new projectile in the same position with the same owner */
      spawn_proj = add_projectile_dmf(round,
				      projectile->munition,
				      projectile->pos, v, NULL),
      set_projectile_owner(spawn_proj, projectile->owner);
      break;
    case MUNITION_EXPLODE:
      calc_explosion(projectile->owner, projectile->pos,
		     action_ptr->parameters[0].float_param,
		     action_ptr->parameters[1].float_param);
      break;
    case MUNITION_CLIP_TO_COLLISION:
      /* If there actually was a collision */
      if(projectile->in_collision)
      {
	/* Swap the position and collision vectors */
	t = projectile->pos;
	projectile->pos = projectile->collision_pos;
	projectile->collision_pos = projectile->pos;

	/* No collision */
	projectile->in_collision = FALSE;
      }
      break;
    case MUNITION_CHANGE_TYPE:
      projectile->munition = find_munition(projectile->munition,
					   action_ptr->parameters[0].string_param);
      break;
    case MUNITION_SET_STATE_RISING:
      projectile->state = RISING;
      break;
    case MUNITION_SET_STATE_FALLING:
      projectile->state = FALLING;
      break;
    case MUNITION_SET_THRUST:
      projectile->thrust = action_ptr->parameters[0].float_param;
      break;
    case MUNITION_SET_DRAG:
      projectile->drag = action_ptr->parameters[0].float_param;
      break;
    case MUNITION_DEPLOY_PARACHUTE:
      projectile->parachute_deployed = TRUE;
      projectile->parachute_drag = action_ptr->parameters[0].float_param;
      break;
    case MUNITION_DROP_PARACHUTE:
      projectile->parachute_deployed = FALSE;
      break;
    case MUNITION_SET_TIMER:
      projectile->timer = action_ptr->parameters[0].float_param;
      projectile->old_timer = projectile->timer - 1.0;
      break;
    case MUNITION_RESET_TIMER:
      projectile->old_timer = -1.0;
      projectile->timer = 0.0;
      break;
    case MUNITION_SPAWN_NEW_CLONE_SPREAD:
    case MUNITION_SPAWN_NEW_SPREAD:
      Neprintf("This action not yet supported!!!! (%d)\n", action_ptr->type);
      break;
    }
    action_ptr = action_ptr->next;
  }

  /* Projectile is still around, so return FALSE (is dead) */
  return FALSE;
}

/* ******************************************* */
/*           PROJECTILE DRAWING CODE           */
/* ******************************************* */

/* Draw all the projectiles in the world */
void draw_projectiles(Round *round)
{
  Map *map = round->map;
  Projectile *proj_ptr = round->projectile_list;
  float c[3] = {0.0, 1.0, 0.0};

  /* Black out all the areas that were lit */
  while(proj_ptr)
  {
    clear_dynamic_light(map, proj_ptr->old_pos, 50.0);

    proj_ptr = proj_ptr->next;
  }

  proj_ptr = round->projectile_list;
  while(proj_ptr)
  {
    draw_projectile(proj_ptr);
    add_dynamic_light(map, proj_ptr->pos, c, 50.0);

    proj_ptr = proj_ptr->next;
  }
}

/* Draw a projectile */
static void draw_projectile(Projectile *projectile)
{
  glPushMatrix();

  glTranslatef(projectile->pos[0],
	       projectile->pos[1],
	       projectile->pos[2]);
  glRotatef(projectile->rot[0], 0.0f, 1.0f, 0.0f);
  glRotatef(projectile->rot[1], 1.0f, 0.0f, 0.0f);
  glRotatef(projectile->rot[2], 0.0f, 0.0f, 1.0f);

  switch(projectile->munition->type)
  {
  default:
  case MUNITION_SHELL:
    glCallList(projectile->world->lists.shell);
    break;
  case MUNITION_MISSILE:
    glCallList(projectile->world->lists.missile);
    break;
  case MUNITION_BOMBLET:
    glCallList(projectile->world->lists.bomblet);
    break;
  }

  /* Draw a parachute if the projectile has one deployed */
  if(projectile->parachute_deployed)
    glCallList(projectile->world->lists.parachute);

  glPopMatrix();
}

/* Remove all projectiles */
int clear_projectiles(Round *round)
{
  Projectile *projectile, *tmp;

  if(!round)
    return FALSE;

  projectile = round->projectile_list;
  while(projectile)
  {
    tmp = projectile->next;
    Nprintf("Removing projectile (Clearing out projectiles)\n");
    remove_projectile(projectile);
    projectile = tmp;
  }

  return TRUE;
}

/* Free a projectile */
static int free_projectile(Projectile *projectile)
{
  if(!projectile)
    return FALSE;

  Nfree(projectile);

  return TRUE;
}
