/*
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __projectile_h__
#define __projectile_h__

typedef enum   _ProjectileState ProjectileState;
typedef struct _Projectile Projectile;

#include "round.h"
#include "player.h"
#include "munition.h"

/* The time in seconds between adding trails */
/* TODO: Make this variable */
#define TRAIL_TIME 0.2

enum _ProjectileState
{
  RISING   = 1,
  FALLING  = 2
};

/* A projectile */
struct _Projectile
{
  World *world;
  Game *game;
  Round *round;

  Player *owner;

  int parachute_deployed;

  float thrust, drag;
  float parachute_drag;

  /* Pointers to the arrays, so can swap two arrays
     faster than copying the values */
  float *pos, *old_pos, *collision_pos;

  /* Flag for whether there is a collision */
  int in_collision;

  /* The actual positions and collision
     (These will be swapped around) */
  float p0[3], p1[3], p2[3];

  /* The up direction of the projectile */
  float up[3];

  /* The position and up-vector of the last smoke trail */
  float old_trail_pos[3], old_trail_up[3], old_trail_left[3];;

  /* The time (in seconds) since the last trail was added */
  float trail_time;

  /* The displacement, velocity, acceleration, and rotation
     vectors for the projectile */
  float dir[3], vel[3], acc[3], rot[3];

  float old_timer, timer, parachute_timer;

  ProjectileState state;
  Munition *munition;

  Projectile *last;
  Projectile *next;
};

/* Funcion prototypes */

/* Projectile add/remove code */

/* Add a projectile based on three vectors, position (displacement),
   velocity (momentum), and acceleration (force) */
Projectile *add_projectile_dmf(Round *round,
			       Munition *munition,
			       float *pos,
			       float *vel,
			       float *acc);

/* Add a projectile based on two vectors, position and direction, and two
   scalars: linear speed and acceleration */
Projectile *add_projectile_pdsa(Round *round,
				Munition *munition,
				float *pos,
				float *dir,
				float spd,
				float thrust);

/* Add a projectile with all it's requirements, type, pos[], vel[], dir[], and acc
   Either the direction or the velocity can be sent as NULL, they can be created from
   eachother */
Projectile *add_projectile(Round *round,
			   Munition *munition,
			   float *pos,
			   float *vel,
			   float *dir,
			   float thrust);

/* Set the owner of the projectile */
void set_projectile_owner(Projectile *projectile, Player *owner);

/* Bye-bye projectile */
void remove_projectile(Projectile *projectile);

/* Slows down time if weapons that need a minimum accuracy are present */
void update_max_dt(Round *round);

/* Projectile update code */
void update_projectiles(Round *round, float);

/* Projectile drawing code */
void draw_projectiles(Round *round);

/* Remove all projectiles */
int clear_projectiles(Round *round);

#endif /* __projectile_h__ */
