/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* trails.c
   This file contains the functions dealing with smoke trails (eg from rockets) */

#include <stdio.h>
#include <malloc.h>
#include <math.h>

#include "trails.h"
#include "game.h"
#include "tools.h"
#include "vector.h"

static int free_trail(Trail *trail);

/* Add a trail */
Trail *add_trail(Round *round, float *pos0, float *pos1,
		 float *up0, float *up1,
		 float *left0, float *left1,
		 GLfloat *colour,
		 float decay_factor, float dt)
{
  Trail *new_trail = malloc(sizeof(Trail));
  int i;

  if(!new_trail)
  {
    fprintf(stderr, "Error: Unable to allocate space for trail");
    return NULL;
  }

  new_trail->world = round->game->world;
  new_trail->game = round->game;
  new_trail->round = round;

  /* Add the trail information in */
  for(i = 0; i < 3; i++)
  {
    new_trail->pos[0][i] = pos0[i];
    new_trail->pos[1][i] = pos1[i];
    new_trail->up[0][i] = up0[i];
    new_trail->up[1][i] = up1[i];
    new_trail->left[0][i] = left0[i];
    new_trail->left[1][i] = left1[i];
  }

  if(colour)
    for(i = 0; i < 4; i++)
      new_trail->colour[i] = colour[i];
  else /* default colour */
  {
    for(i = 0; i < 3; i++)
      new_trail->colour[i] = 0.5;
    new_trail->colour[3] = 1.0;
  }

  new_trail->decay_factor = decay_factor;
  new_trail->decay0 = exp(-new_trail->decay_factor * dt);
  new_trail->decay1 = 1.0;

  new_trail->last = NULL;
  new_trail->next = NULL;

  /* Stick to front of list (so if added wont get acted on straight away) */
  new_trail->next = round->trail_list;
  if(round->trail_list)
    round->trail_list->last = new_trail;
  round->trail_list = new_trail;

  return new_trail;
}

/* Remove a smoke trail */
void remove_trail(Trail *trail)
{
  if(trail->next != NULL)
    trail->next->last = trail->last;
  if(trail->last != NULL)
    trail->last->next = trail->next;
  else
    trail->round->trail_list = trail->next;

  free_trail(trail);
}

/* Make the smoke blow away */
void update_trails(Round *round, float dt)
{
  Trail *trail_ptr = round->trail_list, *temp_trail;

  while(trail_ptr)
  {
    /* Trail may be removed, so update now */
    temp_trail = trail_ptr->next;

    update_trail(trail_ptr, dt);

    trail_ptr = temp_trail;
  }
}

/* Update one trail, removing if necessary */
void update_trail(Trail *trail, float dt)
{
  float decay = exp(-trail->decay_factor * dt);

  /* Make the trail fade away */
  trail->decay0 *= decay;
  trail->decay1 *= decay;
  trail->colour[3] *= decay;

  /* If too faint */
  if(trail->decay0 < 0.01)
    remove_trail(trail);
}

/* Draw smoke trails in the world */
void draw_trails(Round *round)
{
  Trail *trail_ptr = round->trail_list;

  glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT);

  glDisable(GL_TEXTURE_2D);
  glDepthMask(GL_FALSE);
  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_CULL_FACE);

  glBegin(GL_QUADS);
  while(trail_ptr)
  {
    draw_trail(trail_ptr);
    trail_ptr = trail_ptr->next;
  }
  glEnd();

  glPopAttrib();
}

/* Draw one segment of a smoke trail */
void draw_trail(Trail *trail)
{
  float *p0, *p1, up0[3], up1[3], left0[3], left1[3], w0, w1;

  glColor4fv(trail->colour);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, trail->colour);

  p0 = trail->pos[0];
  p1 = trail->pos[1];
  w0 = 10.0 * (1.0 - trail->decay0);
  w1 = 10.0 * (1.0 - trail->decay1);

  _vector_dup(trail->up[0], up0);
  _vector_dup(trail->up[1], up1);
  _vector_scale(w0, up0, up0);
  _vector_scale(w1, up1, up1);

  _vector_dup(trail->left[0], left0);
  _vector_dup(trail->left[1], left1);
  _vector_scale(w0, left0, left0);
  _vector_scale(w1, left1, left1);

  glColor4fv(trail->colour);
  glVertex3f(p0[0] - up0[0], p0[1] - up0[1], p0[2] - up0[2]);
  glVertex3f(p1[0] - up1[0], p1[1] - up1[1], p1[2] - up1[2]);
  glColor4f(0.0, 0.0, 0.0, 0.0);
  glVertex3f(p1[0] + up1[0], p1[1] + up1[1], p1[2] + up1[2]);
  glVertex3f(p0[0] + up0[0], p0[1] + up0[1], p0[2] + up0[2]);

  glColor4fv(trail->colour);
  glVertex3f(p0[0] - left0[0], p0[1] - left0[1], p0[2] - left0[2]);
  glVertex3f(p1[0] - left1[0], p1[1] - left1[1], p1[2] - left1[2]);
  glColor4f(0.0, 0.0, 0.0, 0.0);
  glVertex3f(p1[0] + left1[0], p1[1] + left1[1], p1[2] + left1[2]);
  glVertex3f(p0[0] + left0[0], p0[1] + left0[1], p0[2] + left0[2]);
}

/* Remove all trails */
int clear_trails(Round *round)
{
  Trail *trail, *tmp;

  if(!round)
    return FALSE;

  trail = round->trail_list;
  while(trail)
  {
    tmp = trail->next;
    remove_trail(trail);
    trail = tmp;
  }

  return TRUE;
}

/* Free a trail */
static int free_trail(Trail *trail)
{
  if(!trail)
    return FALSE;

  Nfree(trail);

  return TRUE;
}
