/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* update.c
   Contains the loop function, and calls the functions that update
   dynamic things in the world */

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "update.h"
#include "world.h"
#include "game.h"
#include "camera.h"
#include "round.h"
#include "projectile.h"
#include "trails.h"
#include "player.h"
#include "objects.h"
#include "ai.h"
#include "explosion.h"
#include "satellite.h"
#include "shade.h"
#include "glut_interface.h"
#include "display_stack.h"

/* YP: value which blinks */
float blink_value=0.;

/* counters management */
double counters[MAX_COUNTERS]={0.};
struct timeval ref_tpf={0,12500};  /* 80 fps (freq. of most monitor) */
struct timeval cur_tpf={0,12500};

/* start a counter (given by its number) */
void reset_counter(int c)
{
  if (c >= MAX_COUNTERS)
    return;
  counters[c] = 0.;
}

/* give the elapsed time for the given counter */
double get_counter(int c)
{
  if (c >= MAX_COUNTERS)
    return(0.);
  return(counters[c]);
}

/* set the min. time per frame in microsec (to save computation time) */
void set_min_tpf(int usec)
{
  cur_tpf.tv_usec = usec;
  cur_tpf.tv_sec  = 0;
}

/* reset frame rate to its default value */
void reset_min_tpf()
{
  cur_tpf = ref_tpf;
}

/* Update the dynamic objects in the world */
void update_world(World *world, float dt)
{
  static float fps_dt = 0.0;
  static int dir=1; /* direction for blinking values */
  struct timeval tv;
  int i, *button_states;

  /* Increase total elasped time */
  world->total_elapsed_time += dt;
  /* Increase total elasped time for trigo */
  world->total_elapsed_time_trigo += dt;
  if (world->total_elapsed_time_trigo > 2.*3.14159)
    world->total_elapsed_time_trigo -= 2.*3.14159;

  /* Check the elapsed time and wait if it is smaller
     that the requested time per frame */
  if(dt < cur_tpf.tv_usec / 1000000.0)
  {
    tv.tv_sec = 0;
    tv.tv_usec = (long)((cur_tpf.tv_usec / 1000000.0 - dt)*1000000.0);
    select(0, NULL, NULL, NULL, &tv);
    dt = cur_tpf.tv_usec/1000000.0; /* effective elapsed time */
  }

  /* Manage time varying values */
  blink_value += dir*dt;
  if(dir > 0)
  {
    if(blink_value > 0.99)
    {
      blink_value = 1.;
      dir = -1;
    }
  }
  else
  {
    if(blink_value < 0.01)
    {
      blink_value = 0.;
      dir = 1;
    }
  }

  /* Slow motion */
  /*dt /= 2.0f;*/

  /* manage shading */
  if (world->shade_mode == NATOW_SHADE_DOWN)
    {
    if (update_shade_down(world, dt))
      {
      world->shade_mode = NATOW_SHADE_UP;
      }
    }
  else if (world->shade_mode == NATOW_SHADE_UP)
    {
    if (update_shade_up(world, dt))
      {
      world->shade_mode = NATOW_SHADE_NONE;
      }
    }

  /* a shading is running, do not update world! */
  if(world->shade_mode == NATOW_SHADE_DOWN)
  {
    post_redisplay();
    return;
  }

  /* if paused mode due to cursor, and cursor is inside,
       stop paused mode */
  if ((get_display_mode(world) == NATOW_DISPLAY_ROUND_PAUSED)&&
      (world->paused_from_cursor)&&(world->cursor_inside))
  {
    pop_display_mode(world);
    world->paused_from_cursor = FALSE;
    return;
  }

  /* When not in game, dont update world */
  if(get_display_mode(world) != NATOW_DISPLAY_ROUND)
  {
    post_redisplay();
    return;
  }

  /* If in game but cursor is outside, switch to 'paused' mode */
  if(!world->cursor_inside)
  {
    button_states = get_button_states();

    /* Don't pause if rotating view */
    if((button_states[LEFT_BUTTON] != MOUSE_DOWN) &&
       (button_states[RIGHT_BUTTON] != MOUSE_DOWN))
    {
      push_display_mode(world, NATOW_DISPLAY_ROUND_PAUSED);
      world->paused_from_cursor = TRUE;
      return;
    }
  }

  /* update moons and sun */
  if(world->sun != NULL)
    update_satellite(world->sun, dt);
  if(world->moon1 != NULL)
    update_satellite(world->moon1, dt);
  if(world->moon2 != NULL)
    update_satellite(world->moon2, dt);

  /* increase user counters (only if not paused) */
  for(i = 0; i < MAX_COUNTERS; i++)
    counters[i] += dt;

  if(fps_dt > 1.0f)
  {
    world->fps = 1.0 / dt;
    fps_dt = 0.0f;
  } else
    fps_dt += dt;

  if(dt > world->max_dt)
    dt = world->max_dt;

  /* If a round is being played, update it */
  if(world->game && world->game->round && world->game->mode == TURN_BASED)
    update_round(world->game->round, dt);

  /* in single player mode, update all CPU players */
  if(world->game->mode == SINGLE_PLAYER)
  {
    update_round(world->game->round, dt);
    update_ai_players(world, dt);
  }

  /* Things have changed, so redisplay */
  post_redisplay();
}
